#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
/* #include <getopt.h> */
#include "assembler-const.h"
#include "assembler-yacc.tab.h"

extern int LineNo;
extern char LineBuf[LINEBUFSIZE];
extern int TokPos;
extern FILE *yyin;
extern int Address;
extern char tag[LABELSIZE];
extern int include_stack_ptr;
extern int InitFlag;
extern char currentfilename[256];

int yyrestart(FILE *);
int yyparse(void);

extern struct labeltabletype
{
  char name[LABELSIZE];
  int value;
} labeltable[MAXLABLES];

int Pass;
FILE *symout,*binout;


int yyerror(char *errmsg)
{
  int i,j,t;
  char TempBuf[TABWIDTH*LINEBUFSIZE];
  i=0;
  j=0;
  while (LineBuf[i]!='\0')
    {
      TempBuf[j]=LineBuf[i];
      if (LineBuf[i]=='\t')
	{
	  for (t=0;t<TABWIDTH;t++)
	    TempBuf[j++]=' ';
	  j--;
	}
      i++;
      j++;
    }
  TempBuf[j]='\0';
  fprintf(stderr,"In file: %s\n", currentfilename);
  fprintf(stderr,"\nError line %d: %s\n%s\n",LineNo+1,errmsg,TempBuf);
  for (j=1;j<TokPos;j++)
    fprintf(stderr," ");
  fprintf(stderr,"^\n\n");
  exit(1);
}

int yywarning(char *errmsg)
{
  int i,j,t;
  char TempBuf[TABWIDTH*LINEBUFSIZE];
  i=0;
  j=0;
  while (LineBuf[i]!='\0')
    {
      TempBuf[j]=LineBuf[i];
      if (LineBuf[i]=='\t')
	{
	  for (t=0;t<TABWIDTH;t++)
	    TempBuf[j++]=' ';
	  j--;
	}
      i++;
      j++;
    }
  TempBuf[j]='\0';
  fprintf(stderr,"In file: %s\n", currentfilename);
  fprintf(stderr,"\nWarning line %d: %s\n%s\n",LineNo+1,errmsg,TempBuf);
  for (j=1;j<TokPos;j++)
    fprintf(stderr," ");
  fprintf(stderr,"^\n\n");
  return 0;
}

int oerror(char *errmsg)
{
  int i,j,t;
  char TempBuf[TABWIDTH*LINEBUFSIZE];
  i=0;
  j=0;
  while (LineBuf[i]!='\0')
    {
      TempBuf[j]=LineBuf[i];
      if (LineBuf[i]=='\t')
	{
	  for (t=0;t<TABWIDTH;t++)
	    TempBuf[j++]=' ';
	  j--;
	}
      i++;
      j++;
    }
  TempBuf[j]='\0';
  fprintf(stderr,"In file: %s\n", currentfilename);
  fprintf(stderr,"\nError line %d: %s\n%s\n",LineNo+1,errmsg,TempBuf);

  for (j=1;j<TokPos;j++)
    fprintf(stderr," ");
  fprintf(stderr,"^\n\n");
  exit(1);
}


void init(void)
{
  int n,l;
#ifdef DEBUG
  printf("Initializing table to hold lables\n");
#endif
  for (l=0;l<MAXLABLES;l++)
    {
      for (n=0;n<LABELSIZE;n++)
	{
	  labeltable[l].name[n]=0;
	}
      labeltable[l].value=0;
    }
}

int main(int argc, char *argv[])
{
  char infile[256],symfile[256],binfile[256];
  int c,opts=1;
  init();

  while ((c=getopt(argc,argv,"c"))!=EOF)
    {
      opts++;
    }

  if (opts==argc)
    {   
      fprintf(stderr,"Too few arguments\n");
      exit(1);
    }

  sprintf(infile,"%s.asm",argv[optind]); 
  sprintf(currentfilename,"%s.asm",argv[optind]);
 
  if (opts+1<argc)
    {
      sprintf(symfile,"%s.sym",argv[optind+1]);
      sprintf(binfile,"%s.bin",argv[optind+1]);
    }
  else
    {
      sprintf(symfile,"%s.sym",argv[optind]);
      sprintf(binfile,"%s.bin",argv[optind]);
    }

  symout=fopen("/dev/null","w");
  binout=symout;

  yyin=fopen(infile, "r");
  if(!yyin)
    {
      fprintf(stderr,"Could not open inputfile: %s\n",infile);
      exit(1);
    }

#ifdef DEBUG
      printf("First pass\n");
#endif
      include_stack_ptr = 0;
      InitFlag=0;
      Pass=FIRST;
      yyparse();
      fclose(symout);
      fclose(binout);
#ifdef DEBUG
      printf("Second pass\n");
#endif
      Pass=SECOND;
      sprintf(currentfilename,"%s.asm",argv[optind]);
  
      symout=fopen(symfile,"w");
      if(!symout)
	{
	  fprintf(stderr,"Could not open output file: %s\n",symfile);
	  exit(1);
	}
      binout=fopen(binfile,"w");
      if(!symout)
	{
	  fprintf(stderr,"Could not open output file:%s\n",binfile);
	  exit(1);
	}

      include_stack_ptr = 0;
      InitFlag=0;
      Address=0;
      LineNo=0;
      fclose(yyin);
      yyrestart(yyin);
      yyin=fopen(infile, "r");
      yyparse();
/*       printf("\n"); */
      fclose(symout);
      fclose(binout);
  return 0;
}

