#include <stdio.h>
#include "assembler-const.h"

int oerror(char *);

extern FILE* binout;
extern char *regstr(int, char*);
extern int Pass;
extern int TokPos;
char regstring[10];

int IsNotGReg(int reg)
{
  if ((reg>=0) && (reg<NoOfGRegs))
    return 0;
  else
    return 1;
}


char *binstr(char *dest, int value, int length)
{
  int x;
  int v;
  for (x=0;x<length;x++)
    {
      v=1 << (length-x-1);
      if ( value & v )
	dest[x]='1';
      else
	dest[x]='0';
    }
  dest[x]='\0';
  return dest;
}

void ShiftImmediateRange(int value)
{
  if ((value>15) || (value<0))
    {
      fprintf(stderr,
	      "Immediate value (%d) illegal in shift operation\n",value);
      oerror("Immediate value out of range [0;15]");
    }
}

void AddImmediateRange(int value,int dest, int reg1)
{
  if ((value>63) || (value<-63))
    {
      fprintf(stderr,"Immediate value (%d) illegal in add operation\n",value);
      oerror("Immediate value out of range [-63;63]");
    }
  if (IsNotGReg(dest))
    {
      fprintf(stderr,
	      "Register (%s) illegal in add operation\n",
	      regstr(dest,regstring));
      oerror("Destination register invalid\n");
    }
}

/**************************************************
 Handles: addi, "subi", slli and slri
***************************************************/
void makeopcode_rri(int dest, int reg1, int op, int value)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"reg=reg OP imm:     ");
#endif
  switch (op)
    {
    case ADD_calc: 
      AddImmediateRange(value,dest,reg1);
      fprintf(binout,"%s",OP_addi);
      fprintf(binout,"%s",binstr(deststr,reg1,4));
      fprintf(binout,"%s",binstr(deststr,dest,1));
      if (value<0)
	fprintf(binout,"1%s\n",binstr(deststr,-value,6));
      else
	fprintf(binout,"0%s\n",binstr(deststr,value,6));
      break;
    case SUB_calc: 
      /* This operation is encoded by negating the 'value' */
      AddImmediateRange(value,dest,reg1);
      fprintf(binout,"%s",OP_addi);
      fprintf(binout,"%s",binstr(deststr,reg1,4));
      fprintf(binout,"%s",binstr(deststr,dest,1));
      if (value>0)
	fprintf(binout,"1%s\n",binstr(deststr,value,6));
      else
	fprintf(binout,"0%s\n",binstr(deststr,-value,6));
      break;
    case AND_calc: 
      oerror("Cannot AND immediate");
      break;
    case OR_calc: 
      oerror("Cannot OR immediate");
      break;
    case XOR_calc: 
      oerror("Cannot XOR immediate");
      break;
    case SLL_calc: 
      ShiftImmediateRange(value);
      fprintf(binout,"%s",OP_slli);
      fprintf(binout,"%s",binstr(deststr,dest,4));
      fprintf(binout,"%s",binstr(deststr,reg1,4));
      fprintf(binout,"%s\n",binstr(deststr,value,4));
      break;
    case SLR_calc: 
      ShiftImmediateRange(value);
      fprintf(binout,"%s",OP_slri);
      fprintf(binout,"%s",binstr(deststr,dest,4));
      fprintf(binout,"%s",binstr(deststr,reg1,4));
      fprintf(binout,"%s\n",binstr(deststr,value,4));
      break;
    default:
      fprintf(stderr,"Received operation abbreviation %d\n",op);
      oerror("Unknown operation received");
    }
}

/**************************************************
 Handles: add, sub, and, or, xor, sll and slr
***************************************************/
void makeopcode_rrr(int dest, int reg1, int op, int reg2)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"reg=reg OP reg:     ");
#endif
  switch (op)
    {
    case ADD_calc: 
      strcpy(deststr,OP_add); 
      break;
    case SUB_calc: 
      strcpy(deststr,OP_sub);
      break;
    case AND_calc: 
      strcpy(deststr,OP_and); 
      break;
    case OR_calc: 
      strcpy(deststr,OP_or); 
      break;
    case XOR_calc: 
      strcpy(deststr,OP_xor); 
      break;
    case SLL_calc: 
      strcpy(deststr,OP_sll); 
      break;
    case SLR_calc: 
      strcpy(deststr,OP_slr); 
      break;
    default:
      fprintf(stderr,"Received operation abbreviation %d\n",op);
      oerror("Unknown operation received");
    }
  fprintf(binout,"%s",deststr);
  fprintf(binout,"%s",binstr(deststr,dest,4));
  fprintf(binout,"%s",binstr(deststr,reg1,4));
  fprintf(binout,"%s\n",binstr(deststr,reg2,4));
}

void SetRange(int reg, int value)
{
  if ((value>255) || (value<0))
    {
      fprintf(stderr,"Immediate value (%d) illegal in set operation\n",value);
      oerror("Immediate value out of range [0;255]");
    }
}

void makeopcode_set(int dest, int value)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"reg=imm:            ");
#endif
  SetRange(dest, value);

  fprintf(binout,"%s",OP_set);
  fprintf(binout,"%s",binstr(deststr,dest,4));
  fprintf(binout,"%s\n",binstr(deststr,value,8));
}

void makeopcode_load(int destreg, int reg1, int reg2)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"reg=load(reg+reg):  ");
#endif
  fprintf(binout,"%s",OP_load);
  fprintf(binout,"%s",binstr(deststr,destreg,4));
  fprintf(binout,"%s",binstr(deststr,reg1,4));
  fprintf(binout,"%s\n",binstr(deststr,reg2,4));
}

void makeopcode_store(int reg1, int reg2, int srcreg, int op)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"sto (reg+reg)=reg:  ");
#endif
  fprintf(binout,"%s",OP_store);
  fprintf(binout,"%s",binstr(deststr,srcreg,4));
  fprintf(binout,"%s",binstr(deststr,reg1,4));
  fprintf(binout,"%s\n",binstr(deststr,reg2,4));
}

void LoadImmediateRange(int value, int op, int reg1, int reg2)
{
  if (Pass==SECOND)
    if ((value>63) || (value<-63))
      {
	fprintf(stderr,
		"Immediate value (%d) illegal in load operation\n",value);
	oerror("Immediate value out of range [-63;63]");
      }
  if ((op!=SUB_calc) && (op!=ADD_calc))
    oerror("Illegal operation");
  if (IsNotGReg(reg1))
    {
      fprintf(stderr,"Register (%s) illegal\n",regstr(reg1,regstring));
      oerror("Destination register out of range {G0,G1}");
    }
}

void makeopcode_loadi(int destreg, int reg1, int value, int op)
{
  char deststr[16];

  LoadImmediateRange(value,op,destreg,reg1);
#ifdef DEBUG
  fprintf(binout,"reg=load(reg+imm):  ");
#endif
  fprintf(binout,"%s",OP_loadi);
  fprintf(binout,"%s",binstr(deststr,reg1,4));
  fprintf(binout,"%s",binstr(deststr,destreg,1));
  if (((value>=0) && (op==ADD_calc)) || ((value<0) && (op==SUB_calc)))
    fprintf(binout,"0%s\n",binstr(deststr,value,6));
  else
    {
      if (value<0)
	value=-value;
      fprintf(binout,"1%s\n",binstr(deststr,value,6));
    }
}



void StoreImmediateRange(int value, int op, int reg1, int reg2)
{
  if (Pass==SECOND)
  if ((value>63) || (value<-63))
    {
      fprintf(stderr,"Immediate value (%d) illegal in store operation\n",value);
      oerror("Immediate value out of range [-63;63]");
    }
  if ((op!=SUB_calc) && (op!=ADD_calc))
    oerror("Illegal operation in store operation");
  if (IsNotGReg(reg2))
    {
      fprintf(stderr,"Register (%s) illegal in store operation\n",regstr(reg2,regstring));
      oerror("Source register out of range {G0,G1}");
    }
}

void makeopcode_storei(int reg1, int value, int srcreg, int op)
{
  char deststr[16];
  StoreImmediateRange(value,op, reg1, srcreg);
#ifdef DEBUG
  fprintf(binout,"reg=sto (reg+imm):  ");
#endif
  fprintf(binout,"%s",OP_storei);
  fprintf(binout,"%s",binstr(deststr,reg1,4));
  fprintf(binout,"%s",binstr(deststr,srcreg,1));
  if (((value>=0) && (op==ADD_calc)) || ((value<0) && (op==SUB_calc)))
    fprintf(binout,"0%s\n",binstr(deststr,value,6));
  else
    {
      if (value<0)
	value=-value;
      fprintf(binout,"1%s\n",binstr(deststr,value,6));
    }
}

void JumpRange(int value)
{
  if (Pass!=FIRST)
    if ((value>255) || (value<-255))
      {
	fprintf(stderr,"Immediate value (%d) illegal in jump operation\n",value);
	oerror("Immediate value out of range [255;-255]");
      }
}

void makeopcode_jump(int adr)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"jump:               ");
#endif
  JumpRange(adr);
  fprintf(binout,"%s",OP_jump);
  if (adr<0)
    fprintf(binout,"1%s\n",binstr(deststr,-adr,8));
  else
    fprintf(binout,"0%s\n",binstr(deststr,adr,8));
}

void makeopcode_jumplz(int adr)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"jumplz:             ");
#endif
  JumpRange(adr);
  fprintf(binout,"%s",OP_jumplz);
  if (adr<0)
    fprintf(binout,"1%s\n",binstr(deststr,-adr,8));
  else
    fprintf(binout,"0%s\n",binstr(deststr,adr,8));
}

void makeopcode_jumpnz(int adr)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"jumpnz:             ");
#endif
  JumpRange(adr);
  fprintf(binout,"%s",OP_jumpnz);
  if (adr<0)
    fprintf(binout,"1%s\n",binstr(deststr,-adr,8));
  else
    fprintf(binout,"0%s\n",binstr(deststr,adr,8));
}

void CallRange(int value)
{
  if (Pass!=FIRST)
    if ((value>255) || (value<-255))
      {
	fprintf(stderr,
		"Immediate value (%d) illegal in call operation\n",value);
	oerror("Immediate value out of range [255;-255]");
      }
}

void makeopcode_call(int adr)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"call:               ");
#endif
  CallRange(adr);
  fprintf(binout,"%s",OP_call);
  if (adr<0)
    fprintf(binout,"1%s\n",binstr(deststr,-adr,8));
  else
    fprintf(binout,"0%s\n",binstr(deststr,adr,8));
}

void makeopcode_lcall(int adr)
{
  char deststr[18];
#ifdef DEBUG
  fprintf(binout,"lcall:              ");
#endif
  fprintf(binout,"%s0000\n",OP_lcall);
  fprintf(binout,"                    ");
  fprintf(binout,"%s\n",binstr(deststr,adr,16));
}

void makeopcode_popfifo(int dest, int src)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"popfifo:            ");
#endif
  fprintf(binout,"%s",OP_popfifo);
  fprintf(binout,"%s",binstr(deststr,dest,4));
  fprintf(binout,"%s\n",binstr(deststr,src,4));
}

void PopfifoiRange(int value)
{
  if ((value>15) || (value<0))
    {
      fprintf(stderr,"Immediate value (%d) illegal in popfifo\n",value);
      oerror("Immediate value out of range [0;15]");
    }
}

void makeopcode_popfifoi(int dest, int value)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"popfifoi:           ");
#endif
  PopfifoiRange(value);
  fprintf(binout,"%s",OP_popfifoi);
  fprintf(binout,"%s",binstr(deststr,dest,4));
  fprintf(binout,"%s\n",binstr(deststr,value,4));
}

void makeopcode_readfifo(int dest, int src)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"readfifo:           ");
#endif
  fprintf(binout,"%s",OP_readfifo);
  fprintf(binout,"%s",binstr(deststr,dest,4));
  fprintf(binout,"%s\n",binstr(deststr,src,4));
}

void ReadfifoiRange(int value)
{
  if ((value>15) || (value<0))
    {
      fprintf(stderr,"Immediate value (%d) illegal in readfifoi\n",value);
      oerror("Immediate value out of range [0;15]");
    }
}

void makeopcode_readfifoi(int dest, int value)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"readfifoi:          ");
#endif
  ReadfifoiRange(value);
  fprintf(binout,"%s",OP_readfifoi);
  fprintf(binout,"%s",binstr(deststr,dest,4));
  fprintf(binout,"%s\n",binstr(deststr,value,4));
}

/* void makeopcode_shiftfifo(int reg) */
/* { */
/*   char deststr[16]; */
/* #ifdef DEBUG */
/*   fprintf(binout,"shiftfifo:          "); */
/* #endif */
/*   fprintf(binout,"%s ",OP_shiftfifo); */
/*   fprintf(binout,"%s \n",binstr(deststr,reg,4)); */
/* } */

/* void ShiftfifoiRange(int value) */
/* { */
/*   if ((value>15) || (value<0)) */
/*     { */
/*       fprintf(stderr,"Immediate value (%d) illegal in shiftfifo\n",value); */
/*       oerror("Immediate value out of range [0;15]"); */
/*     } */
/* } */

/* void makeopcode_shiftfifoi(int value) */
/* { */
/*   char deststr[16]; */
/* #ifdef DEBUG */
/*   fprintf(binout,"shiftfifoi:         "); */
/* #endif */
/*   ShiftfifoiRange(value); */
/*   fprintf(binout,"%s ",OP_shiftfifoi); */
/*   fprintf(binout,"%s \n",binstr(deststr,value,4)); */
/* } */

void makeopcode_push(int reg)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"push:               ");
#endif
  fprintf(binout,"%s",OP_push);
  fprintf(binout,"%s\n",binstr(deststr,reg,4));
}

void makeopcode_pop(int reg)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"pop:                ");
#endif
  fprintf(binout,"%s",OP_pop);
  fprintf(binout,"%s\n",binstr(deststr,reg,4));
}

void makeopcode_nop(void)
{
#ifdef DEBUG
  fprintf(binout,"nop:                ");
#endif
  fprintf(binout,"%s0000\n",OP_nop);
}

void makeopcode_rti(void)
{
#ifdef DEBUG
  fprintf(binout,"rti:                ");
#endif
  fprintf(binout,"%s0000\n",OP_rti);
}

void makeopcode_resetfifo(void)
{
#ifdef DEBUG
  fprintf(binout,"resetfifo:          ");
#endif
  fprintf(binout,"%s0000\n",OP_resetfifo);
    
}

void makeopcode_inc(int reg)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"inc:                ");
#endif
  fprintf(binout,"%s",OP_inc);
  fprintf(binout,"%s\n",binstr(deststr,reg,4));
}

void makeopcode_dec(int reg)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"dec:                ");
#endif
  fprintf(binout,"%s",OP_dec);
  fprintf(binout,"%s\n",binstr(deststr,reg,4));
}

void ConstantRange(value)
{
  if ((value>65535) || (value<-65536))
    {
      fprintf(stderr,"Constant (%d) illegal\n",value);
      oerror("Constant out of range [-65536;65536]");
    }
}

void makeopcode_constant(int value)
{
  char deststr[18];
  ConstantRange(value);
#ifdef DEBUG
  fprintf(binout,"constant:           ");
#endif
  fprintf(binout,"%s\n",binstr(deststr,value,16));
}

void makeopcode_zeros(int oldadr, int newadr)
{
  char deststr[18];

  for (;oldadr<newadr;oldadr++)
    {
#ifdef DEBUG
      fprintf(binout,"zero:               ");
#endif
      fprintf(binout,"%s\n",binstr(deststr,65535,16));
    }
}

void makeopcode_mask(int dest, int value)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"mask(reg++,imm):    ");
#endif
  if (value>512)
    fprintf(binout,"%s",OP_mask2);
  else
    fprintf(binout,"%s",OP_mask1);
    
  if (Pass==SECOND)
  switch(value)
    {
    case 64 :
      fprintf(binout,"00"); break;
    case 1024:
      fprintf(binout,"00"); break;

    case 128 :
      fprintf(binout,"01"); break;
    case 2048:
      fprintf(binout,"01"); break;

    case 256 :
      fprintf(binout,"10"); break;
    case 4096:
      fprintf(binout,"10"); break;

    case 512 :
      fprintf(binout,"11"); break;
    case 8192:
      fprintf(binout,"11"); break;

    default:
      fprintf(stderr,"Immediate value (%d) illegal in mask operation\n",value);
      oerror("Immediate value out of range {64,128,256,512,1024,2048,4096,8192}");
    }
  
  fprintf(binout,"%s\n",binstr(deststr,dest,4));
}

void EndnodeRange(int s, int x, int y)
{
  if (Pass==SECOND)
    {
      if ((s>3) || (s<0))
	{
	  fprintf(stderr,"First argument illegal\n");
	  oerror("Constant out of range {1,2,3,4}");
	}
      
      if ((x>15) || (x<0))
	{
	  fprintf(stderr,"Second argument illegal\n");
	  oerror("Constant out of range [0;15]");
	}
      
      if ((y>15) || (y<0))
	{
	  fprintf(stderr,"Third argument illegal\n");
	  oerror("Constant out of range [0;15]");
	}
    }
}

void makeopcode_endnode(int s, int x, int y)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"endnode(s,x,y):     ");
#endif
  EndnodeRange(s,x,y);

  fprintf(binout,"0%s",binstr(deststr,s,2));

  fprintf(binout,"00000");
  
  fprintf(binout,"%s",binstr(deststr,x,4));
  fprintf(binout,"%s\n",binstr(deststr,y,4));
}

void JumpnodeRange(int s, int offset)
{
  if (Pass==SECOND)
    {
      if ((s>3) || (s<0))
	{
	  fprintf(stderr,"First argument illegal\n");
	  oerror("Constant out of range {1,2,3,4}");
	}
      
      if ((offset>8191) || (offset<0))
	{
 	  fprintf(stderr,"Second argument illegal (%d)\n",offset); 
 	  oerror("Constant out of range [0;8191]"); 
	}
    }
}

void makeopcode_jumpnode(int s, int offset)
{
  char deststr[16];
#ifdef DEBUG
  fprintf(binout,"jumpnode(s,off):    ");
#endif
  JumpnodeRange(s,offset);

  fprintf(binout,"1%s",binstr(deststr,s,2));

  fprintf(binout,"%s\n",binstr(deststr,offset,13));
}


